@@ -7,7 +7,7 @@ use parent 'Exporter';
 use Text::Trim qw(trim rtrim);
 use XSLoader;
 
-our $VERSION = '0.000062';
+our $VERSION = '0.000063';
 XSLoader::load( __PACKAGE__, $VERSION );
 
 our @EXPORT_OK = qw[];
@@ -95,7 +95,7 @@ engine
 
 =head1 VERSION
 
-Version 0.000062
+Version 0.000063
 
 =head1 SYNOPSIS
 
@@ -212,7 +212,8 @@ slot. Returns a boolean and avoids all JavaScript to Perl value conversions.
 
 Returns a string with the JavaScript type of a given variable.  Possible
 returned values are C<undefined>, C<null>, C<boolean>, C<number>, C<string>,
-C<object>, C<pointer>, C<buffer>, C<lightfunc>.
+C<array>, C<symbol>, C<function>, C<c_function>, C<lightfunc>, C<thread>,
+C<object>, C<pointer>, C<buffer>.
 
 This method returns C<null> for null values, which fixes the long-standing
 JavaScript bug of returning C<object> for null values.
@@ -241,9 +242,6 @@ of JavaScript code:
 
     setTimeout(function() { X(); }, 0);
 
-This function requires that you first evaluate the JavaScrit file
-C<c_eventloop.js>, which is part of the distribution.
-
 =head2 get_stats
 
 Return a hashref with the statistics gathered as a result of creating the XS
